# -*- coding:utf-8 -*-
# Yi+用户系统接入工具, 适用于flask框架

import hashlib
import redis
import rsa
import requests
import sys
import json
import time

from functools import wraps
from flask import request, jsonify, g
from flask import make_response, url_for


if sys.version_info[0] == 2:
    reload(sys)
    sys.setdefaultencoding("utf-8")


_SECRET_KEY_EX_TIME = 3600 * 24
_AUTHS_EX_TIME = 3600 * 24
_HTTP_CODE = 666

_REDIS_STORE = None
_LOGGER = None
_SSO_DOMAIN = None
_ACCESS_KEY = None
_SECRET_KEY = None

_RESP_HEADERS = {"Content-Type": "application/json"}


class RET:
    OK = 0
    DBERR = 4001
    SESSIONERR = 4101
    PARAMERR = 4103
    TOKENERR = 4400
    NOAUTH = 4401
    SIGNERR = 4402
    SERVERERR = 4500


MSG = {
    RET.OK: "成功",
    RET.DBERR: "数据库错误",
    RET.SESSIONERR: "用户未登录",
    RET.PARAMERR: "参数错误",
    RET.TOKENERR: "令牌无效",
    RET.NOAUTH: "当前用户无访问权限",
    RET.SIGNERR: "参数签名不正确",
    RET.SERVERERR: "服务器内部错误",
}


class UMSError(Exception):
    pass

def _code(errcode, msg=None, data=None):
    resp_dict = dict(errcode=errcode, errmsg=MSG[errcode])
    if msg:
        resp_dict["errmsg"] = msg
    if data:
        resp_dict["data"] = data
    return jsonify(resp_dict)


def _init_var(logger, redis_host, redis_port,
              redis_db, sso_domain, access_key,
              secret_key, redis_password=None):

    global _REDIS_STORE
    global _LOGGER
    global _SSO_DOMAIN
    global _ACCESS_KEY
    global _SECRET_KEY

    _LOGGER = logger
    _SSO_DOMAIN = sso_domain
    _ACCESS_KEY = access_key
    _SECRET_KEY = secret_key

    pool = redis.ConnectionPool(
        port=redis_port,
        host=redis_host,
        db=redis_db,
        password=redis_password
    )
    _REDIS_STORE = redis.StrictRedis(connection_pool=pool)


def to_unicode(s):
    if sys.version_info[0] == 3:
        return s.decode("utf-8") if isinstance(s, bytes) else s

    if isinstance(s, str):
        return s.decode("utf-8")
    elif isinstance(s, unicode):
        return s
    else:
        return unicode(s)


def sign(secret_key, request_info):
    request_info = dict(request_info)
    items = list(request_info.items())
    items.sort()

    sign_string = secret_key
    for key, value in items:
        if key == "sign_key" or key == "SIGN_KEY":
            continue
        sign_string += str(to_unicode(key))
        sign_string += str(to_unicode(value))
    sign_string += secret_key

    hash_string = hashlib.md5()
    hash_string.update(sign_string.encode("utf-8"))
    return hash_string.hexdigest().upper()


def get_request_info():
    if request.form:
        return {key: request.form[key] for key in request.form}
    elif to_unicode(request.data):
        return json.loads(to_unicode(request.data))
    else:
        return {key: request.args[key] for key in request.args}


def _is_debug_mode(request_info):
    return "__debug__" in request_info


def _is_sign_mode(request_info):
    return ("sign_key" in request_info or "SIGN_KEY" in request_info) and     \
           ("access_key" in request_info or "ACCESS_KEY" in request_info) and \
           ("timestamp" in request_info or "TIMESTAMP" in request_info)


def _handle_auths(auth_list, request_info):
    auth_name = ""
    else_auth = ""
    auth_dict = {}

    for auth in auth_list:
        if "condition" in auth:
            auth_dict[auth["condition"]] = auth["auth_name"]
        else:
            else_auth = auth["auth_name"]

    for param in request_info.keys():
        if param in auth_dict:
            auth_name = auth_dict[param]
            break

    return else_auth if auth_name == "" else auth_name


def get_user_secret_key(sso_domain, access_key, secret_key, 
                        user_ack=None, uid=None, redis_cli=None,
                        redis_host=None, redis_port=None,
                        redis_db=None, redis_password=None):
    if not redis_cli:
        pool = redis.ConnectionPool(
            port=redis_port,
            host=redis_host,
            db=redis_db,
            password=redis_password
        )
        redis_cli = redis.StrictRedis(connection_pool=pool)

    # 先从redis中获取secret_key
    if user_ack:
        user_info_json = redis_cli.get(str(user_ack))
    else:
        user_info_json = redis_cli.get(str(user_id))

    if not user_info_json:
        # 无缓存数据
        pub_key, pri_key = rsa.newkeys(1024)
        data = {
            "access_key": access_key,
            "timestamp": int(time.time()),
            "pub_key": pub_key.save_pkcs1().decode()
        }
        if user_ack:
            data["user_access_key"] = user_ack
        else:
            data["user_id"] = uid

        sign_key = sign(secret_key, data)
        data["sign_key"] = sign_key
        resp_json_rsa = requests.post(
            sso_domain + "/api/user_secret_key_info",
            data=json.dumps(data), headers=_RESP_HEADERS
        ).content

        if len(resp_json_rsa) < 200:
            # UMS返回错误
            raise UMSError("用户系统返回错误")

        # UMS返回了正确的加密数据(bytes)
        rsa_list = resp_json_rsa.split(b'wangchaoawangchao')
        user_info_json = ""
        for rsa_string in rsa_list:
            content_string = rsa.decrypt(rsa_string, pri_key).decode()
            user_info_json += content_string
        # 设置缓存
        redis_cli.setex(user_ack, _SECRET_KEY_EX_TIME, user_info_json)

    return json.loads(to_unicode(user_info_json))


def get_user_auths(user_id, sso_domain, redis_cli=None,
                   redis_host=None, redis_port=None,
                   redis_db=None, redis_password=None):

    if not redis_cli:
        pool = redis.ConnectionPool(
            port=redis_port,
            host=redis_host,
            db=redis_db,
            password=redis_password
        )
        redis_cli = redis.StrictRedis(connection_pool=pool)

    auths = redis_cli.get("auths_" + user_id)
    if not auths:
        # 没有缓存数据
        url = "{}/api/users/auths?user_id={}".format(sso_domain, user_id)
        res = requests.get(url).content
        result = json.loads(to_unicode(res))
        if result.get("errcode") != 0:
            raise UMSError("用户系统返回错误")
        auths = result.get("data")
        auths = auths if isinstance(auths, list) else to_unicode(auths)

        # 设置缓存数据
        auth_json = json.dumps(auths)
        redis_cli.setex("auths_" + user_id, _AUTHS_EX_TIME, auth_json)
    else:
        auths = json.loads(auths)

    return auths


def login_required(view_func):
    """验证用户登录状态"""
    @wraps(view_func)
    def wrapper(*args, **kwargs):
        request_info = get_request_info()

        if _is_debug_mode(request_info):
            try:
                check_debug_user_url = _SSO_DOMAIN + "/api/debug_users"
                res = requests.get(check_debug_user_url).content
            except Exception:
                _LOGGER.exception("DB error")
                return _code(RET.DBERR), _HTTP_CODE, _RESP_HEADERS

            res = json.loads(to_unicode(res))
            if res.get("errcode") != 0:
                _LOGGER.info("Forbidden")
                return _code(RET.SESSIONERR), _HTTP_CODE, _RESP_HEADERS

            g.user_id = res.get("user_id")
            g.username = res.get("username")
            g.company_id = res.get("company_id")
            g.company_name = res.get("company_name")
            return view_func(*args, **kwargs)

        elif _is_sign_mode(request_info):
            req_user_access_key = request_info.get(
                "access_key", 
                request_info.get("ACCESS_KEY")
            )
            try:
                user_dict = get_user_secret_key(
                    _SSO_DOMAIN, _ACCESS_KEY, _SECRET_KEY,
                    user_ack=req_user_access_key, redis_cli=_REDIS_STORE
                )
            except Exception:
                _LOGGER.exception("查询用户secret_key时出错")
                return _code(RET.SERVERERR)

            # 校验签名
            secret_key = user_dict.get("secret_key")
            req_sign_key = sign(secret_key, request_info)
            client_sign_key = request_info.get(
                "sign_key", 
                request_info.get("SIGN_KEY")
            )
            if req_sign_key != client_sign_key:
                _LOGGER.info("参数签名不正确")
                return _code(RET.SIGNERR)

            g.user_id = to_unicode(user_dict.get("user_id"))
            g.username = to_unicode(user_dict.get("username"))
            g.company_id = to_unicode(user_dict.get("company_id"))
            g.company_name = to_unicode(user_dict.get("company_name"))
            g.config = to_unicode(user_dict.get("user_config"))
            return view_func(*args, **kwargs)

        else:
            token = request.cookies.get("sessionid", "")
            try:
                user_info = _REDIS_STORE.get("session:" + token)
            except Exception:
                _LOGGER.exception("数据库(redis)异常")
                return _code(RET.DBERR), _HTTP_CODE, _RESP_HEADERS

            if not user_info:
                return _code(RET.SESSIONERR), _HTTP_CODE, _RESP_HEADERS

            user_info = json.loads(to_unicode(user_info))
            g.user_id = to_unicode(user_info.get("user_id"))
            g.username = to_unicode(user_info.get("username"))
            g.company_id = to_unicode(user_info.get("company_id"))
            g.company_name = to_unicode(user_info.get("company_name"))
            g.config = to_unicode(user_info.get("user_config"))
            return view_func(*args, **kwargs)                

    return wrapper


def check_auth(auth_list):
    """鉴权装饰器"""
    def func_wrapper(view_func):
        @wraps(view_func)
        def return_wrapper(*args, **kwargs):
            # 先从redis中查询用户的权限列表
            user_id = g.user_id
            try:
                auths = get_user_auths(user_id, _SSO_DOMAIN, 
                                       redis_cli=_REDIS_STORE)
            except Exception:
                _LOGGER.exception("获取用户权限列表时出错")
                return _code(RET.SERVERERR), _HTTP_CODE, _RESP_HEADERS

            request_info = get_request_info()
            auth_name = auth_list if isinstance(auth_list, str) \
                        else _handle_auths(auth_list, request_info)
            if auth_name not in auths:
                _LOGGER.info("当前用户没有访问权限")
                return _code(RET.NOAUTH), _HTTP_CODE, _RESP_HEADERS

            g.auths = json.dumps(auths)
            return view_func(*args, **kwargs)

        return return_wrapper
    return func_wrapper


def sso_init(app, system_domain, system_name, logger, 
             redis_host, redis_port, redis_db, sso_domain,
             access_key, secret_key, redis_password=None):

    _init_var(logger,
              redis_host,
              redis_port,
              redis_db,
              sso_domain,
              access_key,
              secret_key,
              redis_password)

    @app.route("/sessions", methods=["DELETE"])
    def logout():
        token = request.args.get("token", "")
        try:
            _REDIS_STORE.delete("session:" + token)
        except Exception:
            _LOGGER.exception("删除局部会话时redis出错")
            return _code(RET.DBERR)

        return _code(RET.OK)


    @app.route("/sessions", methods=["GET"])
    @login_required
    def get_session():
        user_info = {
            "username": g.username,
            "user_id": g.user_id,
            "company_name": g.company_name,
            "company_id": g.company_id,
            "config": json.loads(g.config)
        }
        return _code(RET.OK, msg="用户已登录", data=user_info)


    @app.route("/tokens/states", methods=["GET"])
    def check_token():
        token = request.args.get("token")
        if not token:
            _LOGGER.error("token参数缺失")
            return _code(RET.PARAMERR)

        check_token_url = "{}/api/{}/states?logout={}&system={}".format(
            sso_domain, token, system_domain + url_for("logout"), system_name
        )
        try:
            resp = requests.get(check_token_url).content
        except Exception:
            _LOGGER.exception("向用户系统注册时出错")
            return _code(RET.SERVERERR)

        resp = json.loads(to_unicode(resp))
        if resp.get("errcode") == 0:
            user_info = resp.get("data").get("user_info")
            expire_time = resp.get("data").get("expire_time")
            _REDIS_STORE.setex("session:" + token, expire_time, user_info)
            user_id = json.loads(user_info).get("user_id")
            response = make_response(json.dumps(dict(
                errcode=0, errmsg="令牌有效", 
                data={"user_id": user_id}
            )))

            response.set_cookie(
                "sessionid",
                token,
                max_age=expire_time,
                httponly=True
            )
            response.headers["Content-Type"] = "application/json"
            return response
            
        elif resp.get("errcode") == RET.NOAUTH:
            _LOGGER.error("用户没有访问该系统的权限")
            return _code(RET.NOAUTH)

        elif resp.get("errcode") == RET.TOKENERR:
            _LOGGER.error("token无效")
            return _code(RET.TOKENERR)

        else:
            _LOGGER.error("向用户系统注册时, 用户系统返回错误")
            return _code(RET.SERVERERR)


    @app.route("/cache/<key>", methods=["DELETE"])
    def clear_cache(key):
        try:
            _REDIS_STORE.delete(key)
        except Exception:
            _LOGGER.exception("清除缓存数据时redis出错")
            return _code(RET.DBERR)

        return _code(RET.OK)


    @app.route("/cache_users", methods=["DELETE"])
    def clear_user_caches():
        request_info = get_request_info()
        keys = request_info.get("keys")
        try:
            _REDIS_STORE.delete(*key)
        except Exception:
            _LOGGER.exception("清除缓存数据时redis出错")
            return _code(RET.DBERR)

        return _code(RET.OK)


    @app.route("/redis/keys", methods=["DELETE"])
    def del_redis_keys():
        request_info = get_request_info()
        target = request_info.get("target", "")

        if target == "emmmiwanttoflushdb":
            keys = _REDIS_STORE.keys(pattern='*')
        elif target == "session":
            keys = _REDIS_STORE.keys(pattern='session*')
        elif target == "auth":
            keys = _REDIS_STORE.keys(pattern='auths_*')
        else:
            return _code(RET.PARAMERR)

        if len(keys) > 0:
            try:
                _REDIS_STORE.delete(*keys)
            except Exception:
                _LOGGER.exception("清除缓存数据时redis出错")
                return _code(RET.DBERR)

        return _code(RET.OK)